/* ctime.c - a Ruby interface to Corporate Time
 * 
 * $Id: ctime.c,v 1.24 2003/01/07 09:07:17 ianmacd Exp $
 *
 * Author: Ian Macdonald <ian@caliban.org>
 *
 * Copyright (C) 2002-2003 Ian Macdonald
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software Foundation,
 *   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <ruby.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Required for ctapi.h */
#ifdef WIN32
#define CAPI_FUNC(a)  __declspec(dllimport) a
#else
#define CAPI_FUNC(a) a
#endif

#ifdef HAVE_CAPI_AUTHENTICATEASSYSOP
#define HAVE_CAPI_25		/* we're using version 2.5 of CAPI API */
#else
#define HAVE_CAPI_20		/* we're using version 2.0 of CAPI API */
#endif

#include <ctapi.h>
#include "event.h"

#define CTIME_VERSION	"0.1.2"
#define MAX_HANDLES	32

/* some debugging macros */
#define rb_ctime_debug0(msg) if (ruby_debug == Qtrue) { \
			       fprintf(stderr, msg); \
			     }
#define rb_ctime_debug1(msg, arg1) if (ruby_debug == Qtrue) { \
				     fprintf(stderr, msg, arg1); \
				    }
#define rb_ctime_debug2(msg, arg1, arg2) if (ruby_debug == Qtrue) { \
					   fprintf(stderr, msg, arg1, arg2); \
					 }

/* template for an MIME wrapper icalendar event */
static const char *MIME_HEADER = {
    "MIME-Version: 1.0\n"
	"Content-Type: text/calendar\n"
	"Content-Transfer-Encoding: quoted-printable\n\n" "%s"
};

static const char *ICAL_TEMPLATE = {
    "BEGIN:VCALENDAR\n"
	"VERSION:2.0\n"
	"PRODID://Steltor//CAPI//EN\n"
	"BEGIN:VEVENT\n"
	"CLASS:PUBLIC\n" "%s" "END:VEVENT\n" "END:VCALENDAR\n"
};

int agenda_count = 0;
const char *error_name = NULL;
CAPISession session = NULL;
CAPIStatus stat = 0;
CAPIStream my_stream = NULL;
CAPIStatus aStat[MAX_HANDLES];
CAPIHandle agenda[MAX_HANDLES];

/* name of this class */
VALUE cCTime;

/* these are exceptions in the CTime class */
VALUE eCTime_AgendaError;
VALUE eCTime_EventError;
VALUE eCTime_DesignationError;
VALUE eCTime_UnknownUserError;

/* CTime.initialize */
static VALUE ct_init(VALUE self, VALUE server, VALUE user, VALUE passwd)
{
    const char *txt = NULL;
    const char *comma = ",";
    const char *comma_sp = ", ";
    VALUE error;
    VALUE capabilities;
    VALUE cap_list;

    rb_ctime_debug2("Logging onto %s as %s...\n", STR2CSTR(server),
		    STR2CSTR(user));
    stat =
	CAPI_Logon(STR2CSTR(user), STR2CSTR(passwd), STR2CSTR(server),
		   CAPI_FLAGS_NONE, &session);

    CAPI_GetStatusString(stat, &error_name);
    rb_iv_set(self, "@error", rb_str_new2(error_name));
    rb_ctime_debug1("Status = %s\n", error_name);

    if (stat == CAPI_STAT_OK) {

	/* list of e-mail addresses pertaining to open agendas */
	rb_iv_set(self, "@agendas", rb_ary_new());

	/* retrieve informaton on capabilities */
	capabilities = rb_hash_new();

	CAPI_GetCapabilities(session, CAPI_CAPAB_AUTH, CAPI_FLAGS_NONE,
			     &txt);
	cap_list = rb_str_split(rb_str_new2(txt), comma);
	rb_hash_aset(capabilities, rb_str_new2("authentication"),
		     cap_list);

	CAPI_GetCapabilities(session, CAPI_CAPAB_COMP, CAPI_FLAGS_NONE,
			     &txt);
	cap_list = rb_str_split(rb_str_new2(txt), comma);
	rb_hash_aset(capabilities, rb_str_new2("compression"), cap_list);

	CAPI_GetCapabilities(session, CAPI_CAPAB_ENCR, CAPI_FLAGS_NONE,
			     &txt);
	cap_list = rb_str_split(rb_str_new2(txt), comma);
	rb_hash_aset(capabilities, rb_str_new2("encryption"), cap_list);

	CAPI_GetCapabilities(session, CAPI_CAPAB_MAXDATE, CAPI_FLAGS_NONE,
			     &txt);
	rb_hash_aset(capabilities, rb_str_new2("max_date"),
		     rb_str_new2(txt));

	CAPI_GetCapabilities(session, CAPI_CAPAB_CAPI_VERSION,
			     CAPI_FLAGS_NONE, &txt);
	rb_hash_aset(capabilities, rb_str_new2("capi_version"),
		     rb_str_new2(txt));

	CAPI_GetCapabilities(session, CAPI_CAPAB_VERSION, CAPI_FLAGS_NONE,
			     &txt);
	rb_hash_aset(capabilities, rb_str_new2("version"),
		     rb_str_new2(txt));

	CAPI_GetCapabilities(session, CAPI_CAPAB_SERVER_VERSION,
			     CAPI_FLAGS_NONE, &txt);
	rb_hash_aset(capabilities, rb_str_new2("server_version"),
		     rb_str_new2(txt));

	CAPI_GetCapabilities(session, CAPI_CAPAB_UNSUPPORTED_ICAL_PROP,
			     CAPI_FLAGS_NONE, &txt);
	cap_list = rb_str_split(rb_str_new2(txt), comma_sp);
	rb_hash_aset(capabilities, rb_str_new2("unsupported_ical_prop"),
		     cap_list);

	CAPI_GetCapabilities(session, CAPI_CAPAB_UNSUPPORTED_ICAL_COMP,
			     CAPI_FLAGS_NONE, &txt);
	cap_list = rb_str_split(rb_str_new2(txt), comma_sp);
	rb_hash_aset(capabilities, rb_str_new2("unsupported_ical_comp"),
		     cap_list);

	rb_iv_set(self, "@capabilities", capabilities);

	return self;

    } else {

	return Qnil;		/* but why don't we actually get nil
				   assigned to our variable when object
				   instantiation fails? Maybe it's because
				   we return control to CTime.new */
    }
}

/* reader function for @error */
static VALUE ct_error(VALUE self)
{
    return rb_iv_get(self, "@error");
}

/* reader function for @agendas */
static VALUE ct_agendas(VALUE self)
{
    return rb_iv_get(self, "@agendas");
}

/* reader function for @capabilities */
static VALUE ct_capabilities(VALUE self)
{
    return rb_iv_get(self, "@capabilities");
}

/* close opened agendas and disconnect from the server */
static VALUE ct_quit(VALUE self)
{
    rb_ctime_debug0("Quitting from server...\n");

    if (session) {
	if (agenda_count > 0) {
	    rb_ctime_debug1("There are %d agendas open\n", agenda_count);

	    rb_ctime_debug0("Destroying handles...\n");
	    stat =
		CAPI_DestroyHandles(session, agenda, agenda_count,
				    CAPI_FLAGS_NONE);
	    CAPI_GetStatusString(stat, &error_name);
	    agenda_count = 0;

	    /* set @error */
	    rb_iv_set(self, "@error", rb_str_new2(error_name));
	    rb_ctime_debug1("Status = %s\n", error_name);
	}

	rb_ctime_debug0("Logging off...\n");
	stat = CAPI_Logoff(&session, CAPI_FLAGS_NONE);
	CAPI_GetStatusString(stat, &error_name);

	/* set @error */
	rb_iv_set(self, "@error", rb_str_new2(error_name));
	rb_ctime_debug1("Status = %s\n", error_name);

	return self;
    }

    return Qnil;
}

/* open a new agenda */
static VALUE ct_open_agenda(VALUE self, VALUE user)
{
    VALUE agendas;

    rb_ctime_debug1("Opening agenda for %s...\n", STR2CSTR(user));
    agenda[agenda_count] = NULL;

    rb_ctime_debug0("Getting handle...\n");
    stat = CAPI_GetHandle(session, STR2CSTR(user), CAPI_FLAGS_NONE,
			  &agenda[agenda_count]);

    CAPI_GetStatusString(stat, &error_name);
    rb_iv_set(self, "@error", rb_str_new2(error_name));
    rb_ctime_debug1("Status = %s\n", error_name);

    if (stat == CAPI_STAT_OK) {

#ifdef HAVE_CAPI_25
	const char *email = NULL;
#else
	char *email = NULL;
#endif

	rb_ctime_debug0("Getting handle info...\n");
	stat =
	    CAPI_HandleInfo(session, agenda[agenda_count],
			    CAPI_HANDLE_MAILTO, &email);

	CAPI_GetStatusString(stat, &error_name);
	rb_iv_set(self, "@error", rb_str_new2(error_name));
	rb_ctime_debug1("Status = %s\n", error_name);

	if (stat != CAPI_STAT_OK)
	    return Qnil;

	agendas = rb_iv_get(self, "@agendas");
	rb_ary_push(agendas, rb_str_new2(email));
	rb_ctime_debug1("Opened agenda for %s\n", email);

	/* increment the number of opened agendas */
	++agenda_count;

	/* return e-mail address */
	rb_ctime_debug1("There are %d open agendas\n", agenda_count);
	return rb_str_new2(email);

    } else if (stat & CAPI_STAT_DATA_USERID) {

	/* no user on the server matched this name */
	rb_raise(eCTime_AgendaError, "Agenda for %s does not exist",
		 STR2CSTR(user));
    }

    /* shouldn't get here very often */
    rb_ctime_debug0("Program should not reach here\n");

    return Qnil;

}

/* close an open agenda */
static VALUE ct_close_agenda(VALUE self, VALUE user_email)
{
    rb_ctime_debug1("Closing agenda for %s...\n", STR2CSTR(user_email));

    if (agenda_count) {
	int found = 0;
	int i, j;

	for (i = 1; i <= agenda_count; ++i) {

#ifdef HAVE_CAPI_25
	    const char *email = NULL;
#else
	    char *email = NULL;
#endif

	    VALUE agendas;

	    rb_ctime_debug0("Getting handle info...\n");
	    stat =
		CAPI_HandleInfo(session, agenda[i - 1], CAPI_HANDLE_MAILTO,
				&email);

	    CAPI_GetStatusString(stat, &error_name);
	    rb_iv_set(self, "@error", rb_str_new2(error_name));
	    rb_ctime_debug1("Status = %s\n", error_name);

	    if (stat != CAPI_STAT_OK)
		return Qnil;

	    if (strcmp(STR2CSTR(user_email), email) == 0) {

		found = 1;
		rb_ctime_debug0("Destroying handles...\n");
		stat =
		    CAPI_DestroyHandles(session, agenda[i - 1], 1,
					CAPI_FLAGS_NONE);
		CAPI_GetStatusString(stat, &error_name);
		rb_iv_set(self, "@error", rb_str_new2(error_name));
		rb_ctime_debug1("Status = %s\n", error_name);

		agendas = rb_iv_get(self, "@agendas");

		/* close up hole in array */
		for (j = i - 1; j < agenda_count; ++j) {
		    agenda[j] = agenda[j + 1];
		}

		/* remove agenda from @agendas */
		for (j = 1; j <= agenda_count; ++j) {
		    if (strcmp
			(email,
			 STR2CSTR(rb_ary_entry(agendas, j - 1))) == 0) {
			rb_ary_delete_at(agendas, j - 1);
			break;
		    }
		}

		--agenda_count;
	    }
	}

	if (!found)
	    rb_raise(eCTime_AgendaError, "No open agenda for %s",
		     STR2CSTR(user_email));

	rb_ctime_debug1("Closed agenda for %s\n", user_email);
	return user_email;

    } else {
	rb_raise(eCTime_AgendaError, "No agendas opened");
    }
}

/* get the events for a given agenda */
static VALUE ct_get_events(VALUE self, VALUE start_time, VALUE end_time,
			   VALUE all_props)
{
#ifdef HAVE_CAPI_25
    const char *iCalData = NULL;
    const char *data_ptr;
#else
    char *iCalData = NULL;
    char *data_ptr;
#endif
    char *props[2] = { "DTSTART", "DURATION" };
    time_t lTime;
    char start[32];
    char end[32];
    int error = 0;
    char buffer[32768];
    char *crlf;
    int length;
    int event_count = 0;
    VALUE event_list;
    VALUE qp;

    rb_ctime_debug0("Preparing to get events...\n");

    if (agenda_count == 0)
	rb_raise(eCTime_AgendaError, "No agendas opened");

    event_list = rb_ary_new();

    time(&lTime);

    strftime(start, sizeof(start), STR2CSTR(start_time), gmtime(&lTime));
    strftime(end, sizeof(end), STR2CSTR(end_time), gmtime(&lTime));

    rb_ctime_debug2("Will fetch events in range %s to %s...\n", start,
		    end);
    if (all_props == Qtrue) {
	rb_ctime_debug0("Fetching all properties...\n");
    } else {
	rb_ctime_debug2("Fetching %s and %s properties...\n", props[0],
			props[1]);
    }

    /* create a memory stream to receive the information */
    rb_ctime_debug0("Creating memory stream...\n");
    stat =
	CAPI_CreateMemoryStream(session, &my_stream, NULL, &iCalData,
				CAPI_FLAGS_NONE);
    CAPI_GetStatusString(stat, &error_name);

    rb_iv_set(self, "@error", rb_str_new2(error_name));
    rb_ctime_debug1("Status = %s\n", error_name);

    /* fetch the information from the server */
    if (stat == CAPI_STAT_OK) {
	rb_ctime_debug0("Fetching...\n");
	if (all_props == Qtrue) {
	    stat =
		CAPI_FetchEventsByRange(session, agenda, agenda_count,
					aStat, CAPI_FLAGS_NONE, start, end,
					NULL, 0, my_stream);
	} else {
	    stat =
		CAPI_FetchEventsByRange(session, agenda, agenda_count,
					aStat, CAPI_FLAGS_NONE, start, end,
					props, 2, my_stream);
	}
	CAPI_GetStatusString(stat, &error_name);

	rb_iv_set(self, "@error", rb_str_new2(error_name));
	rb_ctime_debug1("Status = %s\n", error_name);
    }

    if (stat != CAPI_STAT_OK)
	error = 1;

    /* release the stream */
    rb_ctime_debug0("Destroying streams...\n");
    stat = CAPI_DestroyStreams(session, &my_stream, 1, CAPI_FLAGS_NONE);
    CAPI_GetStatusString(stat, &error_name);

    rb_iv_set(self, "@error", rb_str_new2(error_name));
    rb_ctime_debug1("Status = %s\n", error_name);

    if (error)
	return Qnil;

    data_ptr = iCalData;

    /* parse through the iCalendar buffer */
    rb_ctime_debug0("Parsing iCalendar stream...\n");
    while (*data_ptr != '\0') {

	/* lines are separated by CR/LF, so look for CR */
	crlf = strchr(data_ptr, '\r');

	/* calculate length of line */
	length = crlf - data_ptr;

	/* copy line into buffer */
	strncpy(buffer, data_ptr, length);

	/* terminate it */
	buffer[length] = '\0';

	if (strncmp(buffer, "BEGIN:VEVENT", 12) == 0) {	/* we have a new
							   event */

	    int found = 1;
	    int attribute;
	    VALUE event;
	    extern VALUE cEvent;

	    ++event_count;

	    /* call Event.new, passing it no arguments */
	    event = rb_funcall(cEvent, rb_intern("new"), 0, NULL);

	    /* advance point to start of next line */
	    data_ptr += strlen(buffer) + 2;

	    /* parse until end of event */
	    while (strncmp(buffer, "END:VEVENT", 10) != 0) {

		crlf = strchr(data_ptr, '\r');
		length = crlf - data_ptr;
		strncpy(buffer, data_ptr, length);
		buffer[length] = '\0';

		/* these are the attributes we're interested in */
		if (strncmp(buffer, "UID:", 4) == 0) {
		    attribute = 1;
		} else if (strncmp(buffer, "DTSTART:", 8) == 0) {
		    attribute = 2;
		} else if (strncmp(buffer, "DURATION:", 9) == 0) {
		    attribute = 3;
		} else if (strncmp(buffer, "SUMMARY:", 8) == 0) {
		    attribute = 4;
		} else if (strncmp(buffer, "LOCATION:", 9) == 0) {
		    attribute = 5;
		} else if (strncmp(buffer, "DESCRIPTION:", 12) == 0) {
		    attribute = 6;
		} else {
		    attribute = 0;
		}

		if (attribute != 0) {	/* we found a relevant line */

		    /* advance data pointer to data after colon */
		    data_ptr = strchr(data_ptr, ':') + 1;

		    crlf = strchr(data_ptr, '\r');
		    length = crlf - data_ptr;
		    strncpy(buffer, data_ptr, length);
		    buffer[length] = '\0';

		    /* decode from quoted printable to array */
		    qp = rb_funcall(rb_str_new2(buffer),
				    rb_intern("unpack"), 1,
				    rb_str_new2("M*"));

		    /* copy back to buffer as string */
		    strcpy(buffer, STR2CSTR(rb_ary_to_s(qp)));

		    /* set instance variable of Event object */
		    switch (attribute) {
		    case 1:
			rb_iv_set(event, "@uid", rb_str_new2(buffer));
			break;
		    case 2:
			rb_iv_set(event, "@start", rb_str_new2(buffer));
			break;
		    case 3:
			rb_iv_set(event, "@duration", rb_str_new2(buffer));
			break;
		    case 4:
			rb_iv_set(event, "@summary", rb_str_new2(buffer));
			break;
		    case 5:
			rb_iv_set(event, "@location", rb_str_new2(buffer));
			break;
		    case 6:
			rb_iv_set(event, "@description",
				  rb_str_new2(buffer));
			break;
		    }

		    /* advance pointer to next line */
		    data_ptr += length + 2;

		} else {

		    /* advance pointer to next line */
		    data_ptr = strchr(data_ptr, '\n') + 1;
		}
	    }

	    if (found) {	/* we have a completed object to add to
				   our array */

		/* add this object to the array */
		rb_ary_push(event_list, event);

		/* reset flag */
		found = 0;
	    }

	} else {

	    /* advance pointer to next line */
	    data_ptr = strchr(data_ptr, '\n') + 1;
	}

    }

    /* return the array */
    rb_ctime_debug1("Found %d events\n", event_count);
    return event_list;
}

/* work as a designate */
static VALUE ct_designate(VALUE self, VALUE user)
{
    rb_ctime_debug1("Setting identity to %s\n", user);
    stat = CAPI_SetIdentity(session, STR2CSTR(user), CAPI_FLAGS_NONE);

    if (stat & CAPI_STAT_SECUR_LOGON_AUTH) {
	/* This user has not granted you designate rights! */
	rb_raise(eCTime_AgendaError,
		 "Insufficient designation rights for user %s",
		 STR2CSTR(user));
    } else if (stat & CAPI_STAT_DATA_USERID) {
	/* No user on the server matched this name! */
	rb_raise(eCTime_UnknownUserError, "User %s unknown",
		 STR2CSTR(user));
    }

    CAPI_GetStatusString(stat, &error_name);
    rb_iv_set(self, "@error", rb_str_new2(error_name));
    rb_ctime_debug1("Status = %s\n", error_name);

    if (stat != CAPI_STAT_OK)
	return Qnil;

    return Qtrue;
}

/* create an event */
static VALUE ct_set_event(VALUE self, VALUE event)
{
    char ical_event[1024];
    char new_event[65536];
    char tmp_buffer[32768];
    char uid_buffer[128];
    const char *new_uid = NULL;
    int error = 0;

    rb_ctime_debug0("Creating a new event on the server...");

    if (agenda_count == 0)
	rb_raise(eCTime_AgendaError, "No agendas opened");

    sprintf(new_event, "DTSTART:%s\n",
	    STR2CSTR(rb_iv_get(event, "@start")));

    if (rb_iv_get(event, "@uid") != Qnil) {
	sprintf(tmp_buffer, "UID:%s\n",
		STR2CSTR(rb_iv_get(event, "@uid")));
	strcat(new_event, tmp_buffer);

	sprintf(tmp_buffer, "RECURRENCE-ID:%s\n",
		STR2CSTR(rb_iv_get(event, "@start")));
	strcat(new_event, tmp_buffer);
    }

    sprintf(tmp_buffer, "DURATION:%s\n",
	    STR2CSTR(rb_iv_get(event, "@duration")));
    strcat(new_event, tmp_buffer);

    if (rb_iv_get(event, "@summary") != Qnil) {
	sprintf(tmp_buffer, "SUMMARY:%s\n",
		STR2CSTR(rb_iv_get(event, "@summary")));
	strcat(new_event, tmp_buffer);
    }

    if (rb_iv_get(event, "@location") != Qnil) {
	sprintf(tmp_buffer, "LOCATION:%s\n",
		STR2CSTR(rb_iv_get(event, "@location")));
	strcat(new_event, tmp_buffer);
    }

    if (rb_iv_get(event, "@description") != Qnil) {
	sprintf(tmp_buffer, "DESCRIPTION:%s\n",
		STR2CSTR(rb_iv_get(event, "@description")));
	strcat(new_event, tmp_buffer);
    }

    sprintf(tmp_buffer, ICAL_TEMPLATE, new_event);
    sprintf(ical_event, MIME_HEADER, tmp_buffer);

    /* store the event */
    rb_ctime_debug0("Creating memory stream...\n");
    stat =
	CAPI_CreateMemoryStream(session, &my_stream, ical_event, NULL,
				CAPI_FLAGS_NONE);
    CAPI_GetStatusString(stat, &error_name);
    rb_iv_set(self, "@error", rb_str_new2(error_name));
    rb_ctime_debug1("Status = %s\n", error_name);

    if (stat == CAPI_STAT_OK) {
	rb_ctime_debug0("Storing event...\n");
	stat =
	    CAPI_StoreEvent(session, agenda, agenda_count, aStat,
			    CAPI_STORE_UPDATE, my_stream);

	CAPI_GetStatusString(stat, &error_name);
	rb_iv_set(self, "@error", rb_str_new2(error_name));
	rb_ctime_debug1("Status = %s\n", error_name);
    }

    if (stat == CAPI_STAT_OK) {
	rb_ctime_debug0("Getting event UID...\n");
	stat = CAPI_GetLastStoredUID(session, &new_uid, CAPI_FLAGS_NONE);

	CAPI_GetStatusString(stat, &error_name);
	rb_iv_set(self, "@error", rb_str_new2(error_name));
	rb_ctime_debug1("Status = %s\n", error_name);

    } else {
	error = 1;
    }

    rb_ctime_debug0("Destroying streams...\n");
    stat = CAPI_DestroyStreams(session, &my_stream, 1, CAPI_FLAGS_NONE);
    CAPI_GetStatusString(stat, &error_name);

    rb_iv_set(self, "@error", rb_str_new2(error_name));
    rb_ctime_debug1("Status = %s\n", error_name);

    if (error) {
	rb_raise(eCTime_AgendaError, "%s", error_name);

	return Qnil;

    }

    strcpy(uid_buffer, new_uid);

    rb_ctime_debug1("Got UID %s...\n", uid_buffer);
    return rb_str_new2(uid_buffer);
}

/* delete an event */
static VALUE ct_delete_event(VALUE self, VALUE uid)
{
    rb_ctime_debug0("Deleting event from server...\n");

    if (agenda_count == 0)
	rb_raise(eCTime_AgendaError, "No agendas opened");

    stat =
	CAPI_DeleteEvent(session, agenda, agenda_count, aStat,
			 CAPI_FLAGS_NONE, STR2CSTR(uid), NULL, 0);
    CAPI_GetStatusString(stat, &error_name);

    rb_iv_set(self, "@error", rb_str_new2(error_name));
    rb_ctime_debug1("Status = %s\n", error_name);

    if (stat == CAPI_STAT_DATA_UID_NOTFOUND)
	rb_raise(eCTime_EventError, "No matching event for UID %s",
		 STR2CSTR(uid));

    if (stat != CAPI_STAT_OK)
	return Qnil;

    return Qtrue;
}

/* initialize this class */
void Init_ctime()
{

    /* define the CTime class */
    cCTime = rb_define_class("CTime", rb_cObject);

    /* define the CTime::AgendaError exception */
    eCTime_AgendaError = rb_define_class_under(cCTime, "AgendaError",
					       rb_eRuntimeError);

    /* define the CTime::EventError exception */
    eCTime_EventError = rb_define_class_under(cCTime, "EventError",
					      rb_eRuntimeError);

    /* define the CTime::EventError exception */
    eCTime_DesignationError =
	rb_define_class_under(cCTime, "DesignationError",
			      rb_eRuntimeError);

    /* define the CTime::UnknownUser exception */
    eCTime_UnknownUserError =
	rb_define_class_under(cCTime, "UnknownUserError",
			      rb_eRuntimeError);

    /* define CTime::VERSION constant */
    rb_define_const(cCTime, "VERSION", rb_str_new2(CTIME_VERSION));

    /* define CTime::ALL_PROPERTIES constant */
    rb_define_const(cCTime, "ALL_PROPERTIES", Qtrue);

    /* define CTime::RESTRICTED_PROPERTIES constant */
    rb_define_const(cCTime, "RESTRICTED_PROPERTIES", Qfalse);

    /* define the CTime.initialize method */
    rb_define_method(cCTime, "initialize", ct_init, 3);

    /* define some accessor methods */
    rb_define_method(cCTime, "error", ct_error, 0);
    rb_define_method(cCTime, "agendas", ct_agendas, 0);
    rb_define_method(cCTime, "capabilities", ct_capabilities, 0);

    /* define the CTime#quit method */
    rb_define_method(cCTime, "quit", ct_quit, 0);

    /* define the CTime#open_agenda method */
    rb_define_method(cCTime, "open_agenda", ct_open_agenda, 1);

    /* define the CTime#close_agenda method */
    rb_define_method(cCTime, "close_agenda", ct_close_agenda, 1);

    /* define the CTime#get_events method */
    rb_define_method(cCTime, "get_events", ct_get_events, 3);

    /* define the CTime#designate method */
    rb_define_method(cCTime, "designate", ct_designate, 1);

    /* define the CTime#set_event method */
    rb_define_method(cCTime, "set_event", ct_set_event, 1);

    /* define the CTime#delete_event method */
    rb_define_method(cCTime, "delete_event", ct_delete_event, 1);

    Init_event();

    return;
}
